@@ -1,15 +1,15 @@
 /*
  * Copyright © 2017, 2018, 2019 Christian Persch
  *
- * This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
- * (at your option) any later version.
+ * This programme is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public
+ * License as published by the Free Software Foundation; either
+ * version 3 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This programme is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <https://www.gnu.org/licenses/>.
@@ -33,13 +33,9 @@
 
 #include "debug.h"
 #include "glib-glue.hh"
-#include "libc-glue.hh"
-#include "utf8.hh"
-
-#ifdef WITH_ICU
 #include "icu-decoder.hh"
 #include "icu-glue.hh"
-#endif
+#include "utf8.hh"
 
 using namespace std::literals;
 
@@ -51,80 +47,98 @@ private:
         bool m_quiet{false};
         bool m_statistics{false};
         bool m_utf8{false};
-        int m_buffer_size{16384};
         int m_repeat{1};
-        vte::glib::StringPtr m_charset{};
-        vte::glib::StrvPtr m_filenames{};
+        char* m_charset{nullptr};
+        char** m_filenames{nullptr};
+
+        template<typename T1, typename T2 = T1>
+        class OptionArg {
+        private:
+                T1* m_return_ptr;
+                T2 m_value;
+        public:
+                OptionArg(T1* ptr, T2 v) : m_return_ptr{ptr}, m_value{v} { }
+                ~OptionArg() { *m_return_ptr = m_value; }
+
+                inline constexpr T2* ptr() noexcept { return &m_value; }
+        };
+
+        using BoolArg = OptionArg<bool, gboolean>;
+        using IntArg = OptionArg<int>;
+        using StrArg = OptionArg<char*>;
+        using StrvArg = OptionArg<char**>;
 
 public:
 
         Options() noexcept = default;
         Options(Options const&) = delete;
         Options(Options&&) = delete;
+
+        ~Options() {
+                if (m_filenames != nullptr)
+                        g_strfreev(m_filenames);
+        }
+
         Options& operator=(Options const&) = delete;
         Options& operator=(Options&&) = delete;
-        ~Options() = default;
 
-        inline constexpr bool   benchmark()   const noexcept { return m_benchmark;   }
-        inline constexpr size_t buffer_size() const noexcept { return std::max(m_buffer_size, 1); }
-        inline constexpr bool   codepoints()  const noexcept { return m_codepoints;  }
-        inline constexpr bool   list()        const noexcept { return m_list;        }
-        inline constexpr bool   statistics()  const noexcept { return m_statistics;  }
-        inline constexpr int    quiet()       const noexcept { return m_quiet;       }
-        inline constexpr bool   utf8()        const noexcept { return m_utf8;        }
-        inline constexpr int    repeat()      const noexcept { return m_repeat;      }
-        inline char const* charset()          const noexcept { return m_charset.get();   }
-        inline char const* const* filenames() const noexcept { return m_filenames.get(); }
+        inline constexpr bool benchmark()  const noexcept { return m_benchmark;  }
+        inline constexpr bool codepoints() const noexcept { return m_codepoints; }
+        inline constexpr bool list()       const noexcept { return m_list;       }
+        inline constexpr bool statistics() const noexcept { return m_statistics; }
+        inline constexpr int  quiet()      const noexcept { return m_quiet;      }
+        inline constexpr bool utf8()       const noexcept { return m_utf8;       }
+        inline constexpr int  repeat()     const noexcept { return m_repeat;     }
+        inline constexpr char const* charset()          const noexcept { return m_charset;   }
+        inline constexpr char const* const* filenames() const noexcept { return m_filenames; }
 
         bool parse(int argc,
                    char* argv[],
                    GError** error) noexcept
         {
-                using BoolOption = vte::ValueGetter<bool, gboolean>;
-                using IntOption = vte::ValueGetter<int, int>;
-                using StringOption = vte::ValueGetter<vte::glib::StringPtr, char*, nullptr>;
-                using StrvOption = vte::ValueGetter<vte::glib::StrvPtr, char**, nullptr>;
+                {
+                        auto benchmark = BoolArg{&m_benchmark, false};
+                        auto codepoints = BoolArg{&m_codepoints, false};
+                        auto list = BoolArg{&m_list, false};
+                        auto quiet = BoolArg{&m_quiet, false};
+                        auto statistics = BoolArg{&m_statistics, false};
+                        auto utf8 = BoolArg{&m_utf8, false};
+                        auto repeat = IntArg{&m_repeat, 1};
+                        auto charset = StrArg{&m_charset, nullptr};
+                        auto filenames = StrvArg{&m_filenames, nullptr};
+                        GOptionEntry const entries[] = {
+                                { "benchmark", 'b', 0, G_OPTION_ARG_NONE, benchmark.ptr(),
+                                  "Measure time spent parsing each file", nullptr },
+                                { "codepoints", 'u', 0, G_OPTION_ARG_NONE, codepoints.ptr(),
+                                  "Output unicode code points by number", nullptr },
+                                { "charset", 'f', 0, G_OPTION_ARG_STRING, charset.ptr(),
+                                  "Input charset", "CHARSET" },
+                                { "list-charsets", 'l', 0, G_OPTION_ARG_NONE, list.ptr(),
+                                  "List available charsets", nullptr },
+                                { "quiet", 'q', 0, G_OPTION_ARG_NONE, quiet.ptr(),
+                                  "Suppress output except for statistics and benchmark", nullptr },
+                                { "repeat", 'r', 0, G_OPTION_ARG_INT, repeat.ptr(),
+                                  "Repeat each file COUNT times", "COUNT" },
+                                { "statistics", 's', 0, G_OPTION_ARG_NONE, statistics.ptr(),
+                                  "Output statistics", nullptr },
+                                { "utf-8", '8', 0, G_OPTION_ARG_NONE, utf8.ptr(),
+                                  "UTF-8 input (default)", nullptr },
+                                { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, filenames.ptr(),
+                                  nullptr, nullptr },
+                                { nullptr },
+                        };
 
-                auto benchmark = BoolOption{m_benchmark, false};
-                auto codepoints = BoolOption{m_codepoints, false};
-                auto list = BoolOption{m_list, false};
-                auto quiet = BoolOption{m_quiet, false};
-                auto statistics = BoolOption{m_statistics, false};
-                auto utf8 = BoolOption{m_utf8, false};
-                auto buffer_size = IntOption{m_buffer_size, 16384};
-                auto repeat = IntOption{m_repeat, 1};
-                auto charset = StringOption{m_charset, nullptr};
-                auto filenames = StrvOption{m_filenames, nullptr};
+                        auto context = g_option_context_new("[FILE…] — decoder cat");
+                        g_option_context_set_help_enabled(context, true);
+                        g_option_context_add_main_entries(context, entries, nullptr);
 
-                GOptionEntry const entries[] = {
-                        { "benchmark", 'b', 0, G_OPTION_ARG_NONE, &benchmark,
-                          "Measure time spent parsing each file", nullptr },
-                        { "buffer-size", 'B', 0, G_OPTION_ARG_INT, &buffer_size,
-                          "Buffer size", "SIZE" },
-                        { "codepoints", 'u', 0, G_OPTION_ARG_NONE, &codepoints,
-                          "Output unicode code points by number", nullptr },
-                        { "charset", 'f', 0, G_OPTION_ARG_STRING, &charset,
-                          "Input charset", "CHARSET" },
-                        { "list-charsets", 'l', 0, G_OPTION_ARG_NONE, &list,
-                          "List available charsets", nullptr },
-                        { "quiet", 'q', 0, G_OPTION_ARG_NONE, &quiet,
-                          "Suppress output except for statistics and benchmark", nullptr },
-                        { "repeat", 'r', 0, G_OPTION_ARG_INT, &repeat,
-                          "Repeat each file COUNT times", "COUNT" },
-                        { "statistics", 's', 0, G_OPTION_ARG_NONE, &statistics,
-                          "Output statistics", nullptr },
-                        { "utf-8", '8', 0, G_OPTION_ARG_NONE, &utf8,
-                          "UTF-8 input (default)", nullptr },
-                        { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &filenames,
-                          nullptr, nullptr },
-                        { nullptr },
-                };
+                        auto rv = bool{g_option_context_parse(context, &argc, &argv, error) != false};
+                        g_option_context_free(context);
+                        if (!rv)
+                                return rv;
+                }
 
-                auto context = vte::take_freeable(g_option_context_new("[FILE…] — decoder cat"));
-                g_option_context_set_help_enabled(context.get(), true);
-                g_option_context_add_main_entries(context.get(), entries, nullptr);
-
-                return g_option_context_parse(context.get(), &argc, &argv, error);
+                return true;
         }
 }; // class Options
 
@@ -214,8 +228,6 @@ public:
 
 }; // class Sink
 
-#ifdef WITH_ICU
-
 static std::unique_ptr<vte::base::ICUDecoder>
 make_decoder(Options const& options)
 {
@@ -249,8 +261,6 @@ make_decoder(Options const& options)
         return std::make_unique<vte::base::ICUDecoder>(converter, u32_converter);
 }
 
-#endif /* WITH_ICU */
-
 class Processor {
 private:
         gsize m_input_bytes{0};
@@ -261,19 +271,18 @@ private:
         template<class Functor>
         void
         process_file_utf8(int fd,
-                          Options const& options,
                           Functor& func)
         {
                 auto decoder = vte::base::UTF8Decoder{};
 
-                auto const buffer_size = options.buffer_size();
-                auto buf = g_new0(uint8_t, buffer_size);
+                auto const buf_size = size_t{16384};
+                auto buf = g_new0(uint8_t, buf_size);
 
                 auto start_time = g_get_monotonic_time();
 
                 auto buf_start = size_t{0};
                 for (;;) {
-                        auto len = read(fd, buf + buf_start, buffer_size - buf_start);
+                        auto len = read(fd, buf + buf_start, buf_size - buf_start);
                         if (!len)
                                 break;
                         if (len == -1) {
@@ -320,24 +329,22 @@ private:
                 g_free(buf);
         }
 
-#ifdef WITH_ICU
         template<class Functor>
         void
         process_file_icu(int fd,
-                         Options const& options,
                          vte::base::ICUDecoder* decoder,
                          Functor& func)
         {
                 decoder->reset();
 
-                auto const buffer_size = options.buffer_size();
-                auto buf = g_new0(uint8_t, buffer_size);
+                auto const buf_size = size_t{16384};
+                auto buf = g_new0(uint8_t, buf_size);
 
                 auto start_time = g_get_monotonic_time();
 
                 auto buf_start = size_t{0};
                 while (true) {
-                        auto len = read(fd, buf + buf_start, buffer_size - buf_start);
+                        auto len = read(fd, buf + buf_start, buf_size - buf_start);
                         if (!len) /* EOF */
                                 break;
                         if (len == -1) {
@@ -374,7 +381,7 @@ private:
                 }
 
                 /* Flush remaining output */
-                auto sptr = reinterpret_cast<uint8_t const*>(buf + buffer_size);
+                auto sptr = reinterpret_cast<uint8_t const*>(buf + buf_size);
                 auto result = vte::base::ICUDecoder::Result{};
                 while ((result = decoder->decode(&sptr, true)) == vte::base::ICUDecoder::Result::eSomething) {
                         func(decoder->codepoint());
@@ -386,7 +393,6 @@ private:
 
                 g_free(buf);
         }
-#endif /* WITH_ICU */
 
         template<class Functor>
         bool
@@ -394,7 +400,6 @@ private:
                      Options const& options,
                      Functor& func)
         {
-#ifdef WITH_ICU
                 auto decoder = std::unique_ptr<vte::base::ICUDecoder>{};
                 if (options.charset()) {
                         decoder = make_decoder(options);
@@ -403,22 +408,17 @@ private:
                 }
 
                 assert(decoder != nullptr || options.charset() == nullptr);
-#endif
 
                 for (auto i = 0; i < options.repeat(); ++i) {
                         if (i > 0 && lseek(fd, 0, SEEK_SET) != 0) {
-                                auto errsv = vte::libc::ErrnoSaver{};
-                                g_printerr("Failed to seek: %s\n", g_strerror(errsv));
+                                g_printerr("Failed to seek: %m\n");
                                 return false;
                         }
 
-#ifdef WITH_ICU
                         if (decoder) {
-                                process_file_icu(fd, options, decoder.get(), func);
-                        } else
-#endif
-                        {
-                                process_file_utf8(fd, options, func);
+                                process_file_icu(fd, decoder.get(), func);
+                        } else {
+                                process_file_utf8(fd, func);
                         }
                 }
 
@@ -458,9 +458,7 @@ public:
                                 } else {
                                         fd = ::open(filename, O_RDONLY);
                                         if (fd == -1) {
-                                                auto errsv = vte::libc::ErrnoSaver{};
-                                                g_printerr("Error opening file %s: %s\n",
-                                                           filename, g_strerror(errsv));
+                                                g_printerr("Error opening file %s: %m\n", filename);
                                         }
                                 }
                                 if (fd != -1) {
@@ -528,17 +526,12 @@ main(int argc,
         }
 
         if (options.list()) {
-#ifdef WITH_ICU
                 auto charsets = vte::base::get_icu_charsets(true);
                 for (auto i = 0; charsets[i]; ++i)
                         g_print("%s\n", charsets[i]);
                 g_strfreev(charsets);
 
                 return EXIT_SUCCESS;
-#else
-                g_printerr("ICU support not available.\n");
-                return EXIT_FAILURE;
-#endif
         }
 
         auto rv = bool{};
