@@ -23,10 +23,8 @@
 
 #include <type_traits>
 
-static_assert(std::is_pod<vte::grid::coords>::value, "vte::grid::coords not POD");
 static_assert(sizeof(vte::grid::coords) == 2 * sizeof(long), "vte::grid::coords size wrong");
 
-static_assert(std::is_pod<vte::grid::span>::value, "vte::grid::span not POD");
 static_assert(sizeof(vte::grid::span) == 4 * sizeof(long), "vte::grid::span size wrong");
 
 static_assert(std::is_pod<vte::view::coords>::value, "vte::view::coords not POD");
@@ -104,6 +102,14 @@ vte::grid::coords::to_string() const
         return buf;
 }
 
+char const*
+vte::grid::halfcoords::to_string() const
+{
+        char *buf = debug_get_buf();
+        g_snprintf(buf, DEBUG_STRING_SIZE, "halfgrid[%ld,%ld%c]", row(), halfcolumn().column(), halfcolumn().half() ? 'R' : 'L');
+        return buf;
+}
+
 char const*
 vte::grid::span::to_string() const
 {
@@ -111,7 +117,7 @@ vte::grid::span::to_string() const
                 return "grid[empty]";
 
         char *buf = debug_get_buf();
-        g_snprintf(buf, DEBUG_STRING_SIZE, "grid[%ld,%ld .. %ld,%ld]",
+        g_snprintf(buf, DEBUG_STRING_SIZE, "grid[(%ld,%ld), (%ld,%ld))",
                    start_row(), start_column(), end_row(), end_column());
         return buf;
 }
@@ -144,6 +150,8 @@ vte::color::rgb::to_string() const
 
 #ifdef MAIN
 
+#pragma GCC diagnostic ignored "-Wunused-variable"
+
 #include <glib.h>
 
 using namespace vte::grid;
@@ -201,6 +209,61 @@ test_grid_coords (void)
 #endif
 }
 
+static void
+test_grid_halfcoords (void)
+{
+        /* Default constructor */
+        halfcoords p1;
+
+        /* Construction and assignment */
+
+        halfcoords p2(16, halfcolumn_t(32, 1));
+        g_assert_cmpint(p2.row(), ==, 16);
+        g_assert_cmpint(p2.halfcolumn().column(), ==, 32);
+        g_assert_cmpint(p2.halfcolumn().half(), ==, 1);
+
+        /* Comparision operators */
+
+        halfcoords a (10, halfcolumn_t(20, 1));
+        halfcoords a2(10, halfcolumn_t(20, 1));
+        halfcoords b (10, halfcolumn_t(21, 0));
+        halfcoords c (10, halfcolumn_t(21, 1));
+        halfcoords d (10, halfcolumn_t(22, 0));
+        halfcoords e (11, halfcolumn_t( 5, 0));
+
+        g_assert_true (a <= a2);
+        g_assert_false(a <  a2);
+        g_assert_false(a >  a2);
+        g_assert_true (a >= a2);
+
+        g_assert_true (a <= b);
+        g_assert_true (a <  b);
+        g_assert_false(a >  b);
+        g_assert_false(a >= b);
+
+        g_assert_true (b <= c);
+        g_assert_true (b <  c);
+        g_assert_false(b >  c);
+        g_assert_false(b >= c);
+
+        g_assert_true (c <= d);
+        g_assert_true (c <  d);
+        g_assert_false(c >  d);
+        g_assert_false(c >= d);
+
+        g_assert_true (d <= e);
+        g_assert_true (d <  e);
+        g_assert_false(d >  e);
+        g_assert_false(d >= e);
+
+#ifdef VTE_DEBUG
+        /* to_string() */
+        g_assert_cmpstr(halfcoords(16, 32, 0).to_string(), ==, "halfgrid[16,32L]");
+        g_assert_cmpstr(halfcoords(16, 32, 1).to_string(), ==, "halfgrid[16,32R]");
+#endif
+}
+
+
 static void
 test_grid_span (void)
 {
@@ -254,8 +317,7 @@ test_grid_span (void)
         g_assert_false(s6.contains(coords(16, 15)));
         g_assert_true (s6.contains(coords(16, 16)));
         g_assert_true (s6.contains(coords(16, 31)));
-        g_assert_true (s6.contains(coords(16, 32)));
-        g_assert_false(s6.contains(coords(16, 33)));
+        g_assert_false(s6.contains(coords(16, 32)));
         g_assert_false(s6.contains(coords(17, 15)));
         g_assert_false(s6.contains(coords(17, 16)));
 
@@ -266,8 +328,8 @@ test_grid_span (void)
         g_assert_true (s7.contains(coords(16, 42)));
         g_assert_true (s7.contains(coords(17, 42)));
         g_assert_true (s7.contains(coords(31, 100)));
-        g_assert_true (s7.contains(coords(32, 8)));
-        g_assert_false(s7.contains(coords(32, 9)));
+        g_assert_true (s7.contains(coords(32, 7)));
+        g_assert_false(s7.contains(coords(32, 8)));
         g_assert_false(s7.contains(coords(33, 2)));
 
         span s8(16, 16, 32, 32);
@@ -277,25 +339,32 @@ test_grid_span (void)
         g_assert_false(s8.box_contains(coords(16, 15)));
         g_assert_true (s8.box_contains(coords(16, 16)));
         g_assert_true (s8.box_contains(coords(16, 24)));
-        g_assert_true (s8.box_contains(coords(16, 32)));
-        g_assert_false(s8.box_contains(coords(16, 33)));
+        g_assert_true (s8.box_contains(coords(16, 31)));
+        g_assert_false(s8.box_contains(coords(16, 32)));
         g_assert_false(s8.box_contains(coords(24, 15)));
         g_assert_true (s8.box_contains(coords(24, 16)));
         g_assert_true (s8.box_contains(coords(24, 24)));
-        g_assert_true (s8.box_contains(coords(24, 32)));
-        g_assert_false(s8.box_contains(coords(24, 33)));
+        g_assert_true (s8.box_contains(coords(24, 31)));
+        g_assert_false(s8.box_contains(coords(24, 32)));
         g_assert_false(s8.box_contains(coords(32, 15)));
         g_assert_true (s8.box_contains(coords(32, 16)));
         g_assert_true (s8.box_contains(coords(32, 24)));
-        g_assert_true (s8.box_contains(coords(32, 32)));
-        g_assert_false(s8.box_contains(coords(32, 33)));
+        g_assert_true (s8.box_contains(coords(32, 31)));
+        g_assert_false(s8.box_contains(coords(32, 32)));
         g_assert_false(s8.box_contains(coords(33, 15)));
         g_assert_false(s8.box_contains(coords(33, 24)));
-        g_assert_false(s8.box_contains(coords(3, 42)));
+        g_assert_false(s8.box_contains(coords(33, 42)));
+
+        /* last_row */
+        span s9(16, 16, 32, 0);
+        g_assert_cmpint(s9.last_row(), ==, 31);
+
+        span s10(16, 16, 32, 1);
+        g_assert_cmpint(s10.last_row(), ==, 32);
 
 #ifdef VTE_DEBUG
         /* to_string() */
-        g_assert_cmpstr(vte::grid::span(17, 42, 18, 3).to_string(), ==, "grid[17,42 .. 18,3]");
+        g_assert_cmpstr(vte::grid::span(17, 42, 18, 3).to_string(), ==, "grid[(17,42), (18,3))");
 #endif
 }
 
@@ -375,6 +444,7 @@ main(int argc, char *argv[])
         g_test_init (&argc, &argv, nullptr);
 
         g_test_add_func("/vte/c++/grid/coords", test_grid_coords);
+        g_test_add_func("/vte/c++/grid/halfcoords", test_grid_halfcoords);
         g_test_add_func("/vte/c++/grid/span", test_grid_span);
         g_test_add_func("/vte/c++/color/rgb", test_color_rgb);
         g_test_add_func("/vte/c++/view/coords", test_view_coords);
